



larch --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
larch_server --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
feff6l --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
feff8l --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
xas_viewer --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
gse_mapviewer --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
xrfdisplay --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
